// Adl2_Sample.cpp : Defines the entry point for the console application.
//

///
///  Copyright (c) 2012 Advanced Micro Devices, Inc.

///  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
///  EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
///  WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.

/// \file Adl2_Sample.cpp
/// \brief Sample  application that demonstrates usage of ADL2 APIs in the application that contains multiple uncoordinated ADL transactions.
///
/// ADL2 APIs have been added to ADL to allow creation of the applications that contain multiple clients that call ADL2 APIs in uncoordinated manner without risking
/// to destroy each others context. You can use legacy ADL and newer ADL2 APIs in the same application. The only limitation is that no more than one client that uses
/// ADL APIs can exist at any given time in the same process. There is no limit to number of clients that use ADL2 APIs. This sample demonstrates how to author application
/// that contains multiple ADL2 transactions and single legacy ADL transaction.
/// Author: Ilia Blank

#if defined (LINUX)
#include "include/adl_sdk.h"
//#include "../../include/customer/oem_structures.h"
#include <dlfcn.h>	//dyopen, dlsym, dlclose
#include <stdlib.h>
#include <string.h>	//memeset
#include <unistd.h>	//sleep
#include <QDebug>

#else
#include <windows.h>
#include <tchar.h>
#include "..\..\include\adl_sdk.h"
#endif

#include <stdio.h>

// Definitions of the used function pointers. Add more if you use other ADL APIs. Note that that sample will use mixture of legacy ADL and ADL2 APIs.
typedef int (*ADL_MAIN_CONTROL_CREATE )(ADL_MAIN_MALLOC_CALLBACK, int );
typedef int (*ADL_MAIN_CONTROL_DESTROY )();
typedef int (*ADL2_MAIN_CONTROL_CREATE )(ADL_MAIN_MALLOC_CALLBACK, int, ADL_CONTEXT_HANDLE*);
typedef int (*ADL2_MAIN_CONTROL_DESTROY )( ADL_CONTEXT_HANDLE);
typedef int (*ADL2_ADAPTER_ACTIVE_GET ) (ADL_CONTEXT_HANDLE, int, int* );
typedef int (*ADL2_DISPLAY_MODES_GET )(ADL_CONTEXT_HANDLE, int iAdapterIndex, int iDisplayIndex, int* lpNumModes, ADLMode** lppModes);
typedef int (*ADL2_DISPLAY_POSSIBLEMODE_GET )(ADL_CONTEXT_HANDLE, int iAdapterIndex, int* lpNumModes, ADLMode** lppModes);
typedef int ( *ADL_ADAPTER_NUMBEROFADAPTERS_GET ) ( int* );
typedef int ( *ADL_ADAPTER_ADAPTERINFO_GET ) ( LPAdapterInfo, int );
typedef int ( *ADL_DISPLAY_DISPLAYINFO_GET ) ( int, int *, ADLDisplayInfo **, int );
typedef int ( *ADL_ADAPTER_ACTIVE_GET ) ( int, int * );
typedef int (*ADL_DISPLAY_POSSIBLEMODE_GET )(int iAdapterIndex, int* lpNumModes, ADLMode** lppModes);
typedef int (*ADL_DISPLAY_MODES_GET )(int iAdapterIndex, int iDisplayIndex, int* lpNumModes, ADLMode** lppModes);
typedef int ( *ADL_ADAPTER_ACTIVE_SET )(int, int, int *);
typedef	int	( *ADL_ADAPTER_PRIMARY_SET )(int );
typedef	int	( *ADL_ADAPTER_MODESWITCH )(int);

ADL_MAIN_CONTROL_CREATE				ADL_Main_Control_Create;
ADL_MAIN_CONTROL_DESTROY			ADL_Main_Control_Destroy;
ADL_ADAPTER_NUMBEROFADAPTERS_GET    ADL_Adapter_NumberOfAdapters_Get;
ADL_ADAPTER_ADAPTERINFO_GET         ADL_Adapter_AdapterInfo_Get;
ADL_DISPLAY_DISPLAYINFO_GET         ADL_Display_DisplayInfo_Get;
ADL_ADAPTER_ACTIVE_GET              ADL_Adapter_Active_Get;
ADL_ADAPTER_ACTIVE_SET ADL_Adapter_Active_Set;
ADL_ADAPTER_PRIMARY_SET ADL_Adapter_Primary_Set;
ADL_ADAPTER_MODESWITCH ADL_Adapter_ModeSwitch;
ADL_DISPLAY_POSSIBLEMODE_GET       ADL_Display_PossibleMode_Get;
ADL_DISPLAY_MODES_GET				ADL_Display_Modes_Get;

ADL2_MAIN_CONTROL_CREATE			ADL2_Main_Control_Create;
ADL2_MAIN_CONTROL_DESTROY			ADL2_Main_Control_Destroy;
ADL2_DISPLAY_MODES_GET				ADL2_Display_Modes_Get;
ADL2_ADAPTER_ACTIVE_GET				ADL2_Adapter_Active_Get;
ADL2_DISPLAY_POSSIBLEMODE_GET       ADL2_Display_PossibleMode_Get;

LPAdapterInfo     lpAdapterInfo = NULL;
LPADLDisplayInfo  lpAdlDisplayInfo = NULL;
LPADLMode lpADLMode = NULL;
LPADLMode lpPossibleADLMode = NULL;

void* __stdcall ADL_Main_Memory_Alloc ( int iSize )
{
    void* lpBuffer = malloc ( iSize );
    return lpBuffer;
}

void __stdcall ADL_Main_Memory_Free ( void** lpBuffer )
{
    if ( NULL != *lpBuffer )
    {
        free ( *lpBuffer );
        *lpBuffer = NULL;
    }
}

#define GetProcAddress dlsym


int InitADL ()
{
#if defined (LINUX)
    void *hDLL;		// Handle to .so library
#else
    HINSTANCE hDLL;		// Handle to DLL
#endif

    #if defined (LINUX)
    hDLL = dlopen( "libatiadlxx.so", RTLD_LAZY|RTLD_GLOBAL);
#else
    hDLL = LoadLibrary("atiadlxx.dll");
    if (hDLL == NULL)
        // A 32 bit calling application on 64 bit OS will fail to LoadLIbrary.
        // Try to load the 32 bit library (atiadlxy.dll) instead
        hDLL = LoadLibrary("atiadlxy.dll");
#endif

    if (NULL == hDLL)
    {
        printf("ADL library not found!\n");
        return ADL_ERR;
    }

    ADL_Main_Control_Create = (ADL_MAIN_CONTROL_CREATE) GetProcAddress(hDLL,"ADL_Main_Control_Create");
    ADL_Main_Control_Destroy = (ADL_MAIN_CONTROL_DESTROY) GetProcAddress(hDLL,"ADL_Main_Control_Destroy");
    ADL_Adapter_NumberOfAdapters_Get = (ADL_ADAPTER_NUMBEROFADAPTERS_GET)GetProcAddress(hDLL,"ADL_Adapter_NumberOfAdapters_Get");
    ADL_Adapter_AdapterInfo_Get                     = (ADL_ADAPTER_ADAPTERINFO_GET) GetProcAddress(hDLL,"ADL_Adapter_AdapterInfo_Get");
    ADL_Display_DisplayInfo_Get                     = (ADL_DISPLAY_DISPLAYINFO_GET) GetProcAddress(hDLL,"ADL_Display_DisplayInfo_Get");
    ADL_Adapter_Active_Get                          = (ADL_ADAPTER_ACTIVE_GET)GetProcAddress(hDLL,"ADL_Adapter_Active_Get");

    ADL_Display_PossibleMode_Get = (ADL_DISPLAY_POSSIBLEMODE_GET) GetProcAddress(hDLL,"ADL_Display_PossibleMode_Get");
    ADL_Display_Modes_Get = (ADL_DISPLAY_MODES_GET) GetProcAddress(hDLL,"ADL_Display_Modes_Get");

    ADL2_Main_Control_Create = (ADL2_MAIN_CONTROL_CREATE) GetProcAddress(hDLL,"ADL2_Main_Control_Create");
    ADL2_Main_Control_Destroy = (ADL2_MAIN_CONTROL_DESTROY) GetProcAddress(hDLL,"ADL2_Main_Control_Destroy");
    ADL2_Display_Modes_Get = (ADL2_DISPLAY_MODES_GET) GetProcAddress(hDLL,"ADL2_Display_Modes_Get");
    ADL2_Display_PossibleMode_Get = (ADL2_DISPLAY_POSSIBLEMODE_GET) GetProcAddress(hDLL,"ADL2_Display_PossibleMode_Get");
    ADL2_Adapter_Active_Get = (ADL2_ADAPTER_ACTIVE_GET)GetProcAddress(hDLL, "ADL2_Adapter_Active_Get");

    ADL_Adapter_Active_Set                            = (ADL_ADAPTER_ACTIVE_SET) GetProcAddress(hDLL, "ADL_Adapter_Active_Set");
    ADL_Adapter_ModeSwitch                           = (ADL_ADAPTER_MODESWITCH) GetProcAddress(hDLL, "ADL_Adapter_ModeSwitch");
    ADL_Adapter_Primary_Set                           = (ADL_ADAPTER_PRIMARY_SET) GetProcAddress(hDLL, "ADL_Adapter_Primary_Set");


    if (NULL == ADL_Main_Control_Create ||
        NULL == ADL_Main_Control_Destroy ||
        NULL == ADL_Adapter_NumberOfAdapters_Get ||
        NULL == ADL2_Main_Control_Create ||
        NULL == ADL2_Main_Control_Destroy ||
        NULL == ADL2_Display_Modes_Get ||
        NULL == ADL2_Adapter_Active_Get)
    {
        printf("ADL's API is missing!\n");
        return ADL_ERR;
    }

    if (hDLL != NULL)
    {
        ADL_Main_Control_Create = (ADL_MAIN_CONTROL_CREATE)GetProcAddress(hDLL,"ADL_Main_Control_Create");
        if ( NULL != ADL_Main_Control_Create)
            ADL_Main_Control_Create (ADL_Main_Memory_Alloc, 1);
    }
    else { qDebug() << "Error: ADL_Main_Control_Create"; return 0; }

    return ADL_OK;
}

int AdlModeToFile( LPADLMode lpADLMode, bool possible=false )
{
    if (NULL == lpADLMode ) return ADL_ERR_NULL_POINTER;

    if(possible) {
        printf( "\t\t%-2s %d", "A-idx", lpADLMode->iAdapterIndex );
        printf( " %-2s %d", "D-Idx", lpADLMode->displayID.iDisplayLogicalIndex);
        printf( " %-2s %d", "W", lpADLMode->iXRes);
        printf( " %-2s %d\t", "H", lpADLMode->iYRes);
        printf( " %-2s %.2f\n", "RR", lpADLMode->fRefreshRate);
        return ADL_OK;
    }
    printf( "%-15s %d\n", "AdapterIndex", lpADLMode->iAdapterIndex );
    printf( "%-15s %d\n", "DisplayIndex", lpADLMode->displayID.iDisplayLogicalIndex);
    printf( "%-15s %d\n", "Width", lpADLMode->iXRes);
    printf( "%-15s %d\n", "Height", lpADLMode->iYRes);
    printf( "%-15s %d\n", "ColorDepth", lpADLMode->iColourDepth);
    printf( "%-15s %.2f\n", "RefreshRate", lpADLMode->fRefreshRate);
    printf( "%-15s %d\n", "XPos", lpADLMode->iXPos);
    printf( "%-15s %d\n", "YPos", lpADLMode->iYPos);
    printf( "%-15s %d\n", "Orientation", lpADLMode->iOrientation);
    printf( "%-15s %d\n", "ModeFlag", lpADLMode->iModeFlag);
    printf( "%-15s 0x%04X\n", "ModeMask", lpADLMode->iModeMask);
    printf( "%-15s 0x%04X\n\n", "ModeValue", lpADLMode->iModeValue);

    return ADL_OK;
}

int main (int c,char* v[])
{
    int ADL_Err = InitADL ();

    if (ADL_Err) { return 0; }

    int  iActive;
    int  iNumberAdapters;
    int  iAdapterIndex;
    int  iDisplayIndex;
    int  iNumDisplays;
    int  iNumModes;



    ADL_Adapter_NumberOfAdapters_Get ( &iNumberAdapters );

    if ( 0 < iNumberAdapters )
    {
        lpAdapterInfo = malloc ( sizeof (AdapterInfo) * iNumberAdapters );
        memset ( lpAdapterInfo,'\0', sizeof (AdapterInfo) * iNumberAdapters );

        // Get the AdapterInfo structure for all adapters in the system
        ADL_Adapter_AdapterInfo_Get (lpAdapterInfo, sizeof (AdapterInfo) * iNumberAdapters);
    }
    else { qDebug() << "Error: ADL_Adapter_NumberOfAdapters_Get" << iNumberAdapters; return 0; }

    // Repeat for all available adapters in the system
    for (int i = 0; i < iNumberAdapters; i++ )
    {
         iAdapterIndex = lpAdapterInfo[ i ].iAdapterIndex;

         ADL_Err = ADL_Adapter_Active_Get ( iAdapterIndex, &iActive );
         // If the adapter is not active skip the steps below.
         if ( 0 == iActive || ADL_OK != ADL_Err)
             continue;

         ADL_Main_Memory_Free ( (void **)&lpAdlDisplayInfo );

         ADL_Err = ADL_Display_DisplayInfo_Get (iAdapterIndex, &iNumDisplays, &lpAdlDisplayInfo, 1);

         if (ADL_OK != ADL_Err) continue;

         for (int j = 0; j < iNumDisplays; j++ )
         {
             // For each display, check its status. Use the display only if it's connected AND mapped (iDisplayInfoValue: bit 0 and 1 )
             if (  ( ADL_DISPLAY_DISPLAYINFO_DISPLAYCONNECTED | ADL_DISPLAY_DISPLAYINFO_DISPLAYMAPPED ) !=
                 ( ADL_DISPLAY_DISPLAYINFO_DISPLAYCONNECTED | ADL_DISPLAY_DISPLAYINFO_DISPLAYMAPPED	 &
                     lpAdlDisplayInfo[ j ].iDisplayInfoValue ) )
                     continue;   // Skip the not connected or not mapped displays

                 // Is the display mapped to this adapter?
                 if ( iAdapterIndex != lpAdlDisplayInfo[ j ].displayID.iDisplayLogicalAdapterIndex )
                     continue;

                 iDisplayIndex = lpAdlDisplayInfo[ j ].displayID.iDisplayLogicalIndex;

                 // Obtain the ADL_Display_Modes_Get
                 ADL_Err = ADL_Display_Modes_Get ( iAdapterIndex, iDisplayIndex, &iNumModes, &lpADLMode );
                 if ( ADL_OK != ADL_Err ) { qDebug() << "Error: ADL_Display_Modes_Get"; return 0; }
                 else {
                     if ( lpADLMode ) { ADL_Err = AdlModeToFile( lpADLMode ); }
                     if ( ADL_OK != ADL_Err ) { qDebug() << "Error: AdlModeToFile" ; return 0; }

                       int iNewlyActive;
                       int iStatus = ADL_TRUE;
                       int iRet = ADL_Adapter_ModeSwitch(lpADLMode->iAdapterIndex);

                       lpPossibleADLMode = malloc ( sizeof (LPADLMode) * 13 );
                       memset ( lpPossibleADLMode,'\0', sizeof (LPADLMode) * 13 );

                       ADL_Err = ADL_Display_PossibleMode_Get ( lpADLMode->iAdapterIndex, &iNumModes, &lpPossibleADLMode );
                       if ( ADL_OK != ADL_Err ) { qDebug() << "Error: ADL_Display_PossibleMode_Get"; return 0; }
                       else {
                            for(int i=0; i < iNumModes; ++i) { AdlModeToFile(&lpPossibleADLMode[i],true); }
                       }

                 }
         }
    }
}
